from abc import ABC, abstractmethod

from autograd import value_and_grad, vector_jacobian_product

import autograd.numpy as np
import autograd.numpy.random as npr

__all__ = [
    'VariationalObjective',
    'StochasticVariationalObjective',
    'ExclusiveKL',
    'AlphaDivergence'
]


class VariationalObjective(ABC):
    """A class representing a variational objective to minimize"""
    def __init__(self, approx, model):
        """
        Parameters
        ----------
        approx : `ApproximationFamily` object
        model : `Model` object
        """
        self._approx = approx
        self._model = model
        self._objective_and_grad = None
        self._update_objective_and_grad()


    def __call__(self, var_param):
        """Evaluate objective and its gradient.

        May produce an (un)biased estimate of both.

        Parameters
        ----------
        var_param : `numpy.ndarray`, shape (var_param_dim,)
            The variational parameter.
        """
        if self._objective_and_grad is None:
            raise RuntimeError("no objective and gradient available")
        return self._objective_and_grad(var_param)

    @abstractmethod
    def _update_objective_and_grad(self):
        """Update the objective and gradient function.

        Should be called whenever a parameter that the objective depends on
        (e.g., `approx` or `model`) is updated."""
        pass

    @property
    def approx(self):
        """The approximation family."""
        return self._approx

    @approx.setter
    def approx(self, value):
        self._approx = value
        self._update_objective_and_grad()

    @property
    def model(self):
        """The model."""
        return self._model

    @model.setter
    def model(self, value):
        self._model = value
        self._update_objective_and_grad()


class StochasticVariationalObjective(VariationalObjective):
    """A class representing a variational objective approximated using Monte Carlo."""
    def __init__(self, approx, model, num_mc_samples):
        """
        Parameters
        ----------
        approx : `ApproximationFamily` object
        model : `Model` object
        num_mc_sample : `int`
            Number of Monte Carlo samples to use to approximate the objective.
        """
        self._num_mc_samples = num_mc_samples
        super().__init__(approx, model)

    @property
    def num_mc_samples(self):
        """Number of Monte Carlo samples to use to approximate the objective."""
        return self._num_mc_samples

    @num_mc_samples.setter
    def num_mc_samples(self, value):
        self._num_mc_samples = value
        self._update_objective_and_grad()


class ExclusiveKL(StochasticVariationalObjective):
    """Exclusive Kullback-Leibler divergence.

    Equivalent to using the canonical evidence lower bound (ELBO)
    """
    def _update_objective_and_grad(self):
        approx = self.approx
        def variational_objective(var_param):
            samples = approx.sample(var_param, self.num_mc_samples)
            if approx.supports_entropy:
                lower_bound = np.mean(self.model(samples)) + approx.entropy(var_param)
            else:
                lower_bound = np.mean(self.model(samples) - approx.log_density(samples))
            return -lower_bound
        self._objective_and_grad = value_and_grad(variational_objective)


class AlphaDivergence(StochasticVariationalObjective):
    """Log of the alpha-divergence."""
    def __init__(self, approx, model, num_mc_samples, alpha):
        """
        Parameters
        ----------
        approx : `ApproximationFamily` object
        model : `Model` object
        num_mc_sample : `int`
            Number of Monte Carlo samples to use to approximate the objective.
        alpha : `float`
        """
        self._alpha = alpha
        super().__init__(approx, model, num_mc_samples)

    @property
    def alpha(self):
        """Alpha parameter of the divergence."""
        return self._alpha

    def _update_objective_and_grad(self):
        """Provides a stochastic estimate of the variational lower bound."""
        def compute_log_weights(var_param, seed):
            samples = self.approx.sample(var_param, self.num_mc_samples, seed)
            log_weights = self.model(samples) - self.approx.log_density(var_param, samples)
            return log_weights

        log_weights_vjp = vector_jacobian_product(compute_log_weights)
        alpha = self.alpha
        # manually compute objective and gradient
        def objective_grad_and_log_norm(var_param):
            # must create a shared seed!
            seed = npr.randint(2**32)
            log_weights = compute_log_weights(var_param, seed)
            log_norm = np.max(log_weights)
            scaled_values = np.exp(log_weights - log_norm)**alpha
            obj_value = np.log(np.mean(scaled_values))/alpha + log_norm
            obj_grad = alpha*log_weights_vjp(var_param, seed, scaled_values) / scaled_values.size
            return (obj_value, obj_grad)

        self._objective_and_grad = objective_grad_and_log_norm



import autograd.numpy as np
from autograd.scipy import stats
from autograd.scipy import special
from autograd.numpy import linalg


# See: https://github.com/scipy/scipy/blob/master/scipy/stats/_multivariate.py
def multivariate_t_logpdf(x, m, S, df=np.inf):
    """calculate log pdf for each value

    Parameters
    ----------
    x : array_like, shape=(n_samples, n_features)

    m : array_like, shape=(n_features,)

    S : array_like, shape=(n_features, n_features)
        covariance  matrix
    df : int or float
        degrees of freedom
    """
    #m = np.asarray(m)
    d = m.shape[-1]
    if df == np.inf:
        return stats.multivariate_normal.logpdf(x, m, S)
    #psd = _PSD(S)
    s, u = linalg.eigh(S)
    eps = 1e-10
    s_pinv = np.array([0 if abs(x) <= eps else 1/x for x in s], dtype=float)
    U = np.multiply(u, np.sqrt(s_pinv))
    log_pdet = np.sum(np.log(s))

    log_pdf = special.gammaln(.5*(df + d)) - special.gammaln(.5*df) - .5*d * np.log(np.pi * df)
    log_pdf += -.5*log_pdet
    dev = x - m
    maha = np.sum(np.square(np.dot(dev, U)), axis=-1)
    log_pdf += -.5*(df + d) * np.log(1 + maha / df)
    return log_pdf



import autograd.numpy as np

import tqdm

__all__ = [
    'adagrad_optimize',
]


def learning_rate_schedule(n_iters, learning_rate, learning_rate_end):
    """
    This function sets the schedule for the learning rate,
    the learning rate is kept constant for first 25% and last 25 % iterations,
    but it is linearly reduced from 25-75% iterations from learning_rate to learning_rate_end
    such that the 25th and 75th quantile are given by learning_rate_end and learning_rate respectively.

    Parameters
    ----------
    n_iters : number of iterations
    learning_rate : starting learning rate
    learning_rate_end : ending learning rate

    Returns
    -------
    learning_rate : generator for learning rate
    """

    if learning_rate < 0:
        raise ValueError('learning rate must be positive')
    if learning_rate_end is not None:
        if learning_rate <= learning_rate_end:
            raise ValueError('initial learning rate must be greater than final learning rate')
        # constant learning rate for first quarter, then decay like a/(b + i)
        # for middle half, then constant for last quarter
        b = n_iters*learning_rate_end/(2*(learning_rate - learning_rate_end))
        a = learning_rate*b
        start_decrease_at = n_iters//4
        end_decrease_at = 3*n_iters//4
    for i in range(n_iters):
        if learning_rate_end is None or i < start_decrease_at:
            yield learning_rate
        elif i < end_decrease_at:
            yield a / (b + i - start_decrease_at + 1)
        else:
            yield learning_rate_end


def adagrad_optimize(n_iters, objective, init_param,
                     has_log_norm=False, window=10, learning_rate=.01,
                     epsilon=.1, learning_rate_end=None):
    """
    This is adagrad optimizer without convergence diagnostics, we keep it as a baseline
    , and for cases where the dimensionality is so high that
    the optimizers may become too slow ....

    :param n_iters:
    :param objective:
    :param init_param:
    :param has_log_norm:
    :param window:
    :param learning_rate:
    :param epsilon:
    :param learning_rate_end:
    :return:
    """
    local_grad_history = []
    local_log_norm_history = []
    value_history = []
    log_norm_history = []
    variational_param = init_param.copy()
    variational_param_history = []
    with tqdm.trange(n_iters) as progress:
        try:
            schedule = learning_rate_schedule(n_iters, learning_rate, learning_rate_end)
            for i, curr_learning_rate in zip(progress, schedule):
                prev_variational_param = variational_param
                if has_log_norm:
                    obj_val, obj_grad, log_norm = objective(variational_param)
                else:
                    obj_val, obj_grad = objective(variational_param)
                    log_norm = 0
                value_history.append(obj_val)
                local_grad_history.append(obj_grad)
                local_log_norm_history.append(log_norm)
                log_norm_history.append(log_norm)
                if len(local_grad_history) > window:
                    local_grad_history.pop(0)
                    local_log_norm_history.pop(0)
                grad_scale = np.exp(np.min(local_log_norm_history) - np.array(local_log_norm_history))
                scaled_grads = grad_scale[:,np.newaxis]*np.array(local_grad_history)
                accum_sum = np.sum(scaled_grads**2, axis=0)
                variational_param = variational_param - curr_learning_rate*obj_grad/np.sqrt(epsilon + accum_sum)
                variational_param_history.append(variational_param.copy())
                # retain the last 25% of iterates
                if len(variational_param_history) > i // 4 + 1:
                    variational_param_history.pop()
                if i % 10 == 0:
                    avg_loss = np.mean(value_history[max(0, i - 1000):i + 1])
                    progress.set_description(
                        'Average Loss = {:,.5g}'.format(avg_loss))
        except (KeyboardInterrupt, StopIteration) as e:  # pragma: no cover
            # do not print log on the same line
            progress.close()
        finally:
            progress.close()
    variational_param_history = np.array(variational_param_history)
    smoothed_opt_param = np.mean(variational_param_history, axis=0)
    return (smoothed_opt_param, variational_param_history,
            np.array(value_history), np.array(log_norm_history))



from abc import ABC, abstractmethod

import autograd.numpy as np
import autograd.numpy.random as npr
import autograd.scipy.stats.multivariate_normal as mvn
import autograd.scipy.stats.t as t_dist
from autograd.scipy.linalg import sqrtm
from scipy.linalg import eigvalsh

from paragami import (PatternDict,
                      NumericVectorPattern,
                      PSDSymmetricMatrixPattern,
                      FlattenFunctionInput)

from ._distributions import multivariate_t_logpdf

__all__ = [
    'ApproximationFamily',
    'MFGaussian',
    'MFStudentT',
    'MultivariateT',
]

class ApproximationFamily(ABC):
    """An abstract class for an variational approximation family.

    See derived classes for examples.
    """
    def __init__(self, dim, var_param_dim, supports_entropy, supports_kl):
        """
        Parameters
        ----------
        dim : `int`
            The dimension of the space the distributions in the approximation family are defined on.
        var_param_dim : `int`
            The dimension of the variational parameter.
        supports_entropy : `bool`
            Whether the approximation family supports closed-form entropy computation.
        supports_kl : `bool`
            Whether the approximation family supports closed-form KL divergence
            computation.
        """
        self._dim = dim
        self._var_param_dim = var_param_dim
        self._supports_entropy = supports_entropy
        self._supports_kl = supports_kl

    def init_param(self):
        """A variational parameter to use for initialization.

        Returns
        -------
        var_param : `numpy.ndarray`, shape (var_param_dim,)
        """
        return np.zeros(self.var_param_dim)

    @abstractmethod
    def sample(self, var_param, n_samples, seed=None):
        """Generate samples from the variational distribution

        Parameters
        ----------
        var_param : `numpy.ndarray`, shape (var_param_dim,)
            The variational parameter.
        n_samples : `int`
            The number of samples to generate.

        Returns
        -------
        samples : `numpy.ndarray`, shape (n_samples, var_param_dim)
        """
        pass

    def entropy(self, var_param):
        """Compute entropy of variational distribution.

        Parameters
        ----------
        var_param : `numpy.ndarray`, shape (var_param_dim,)
            The variational parameter.

        Raises
        ------
        NotImplementedError
            If entropy computation is not supported."""
        if self.supports_entropy:
            return self._entropy(var_param)
        raise NotImplementedError()

    def _entropy(self, var_param):
        raise NotImplementedError()

    @property
    def supports_entropy(self):
        """Whether the approximation family supports closed-form entropy computation."""
        return self._supports_entropy

    def kl(self, var_param0, var_param1):
        """Compute the Kullback-Leibler (KL) divergence.

        Parameters
        ----------
        var_param0, var_param1 : `numpy.ndarray`, shape (var_param_dim,)
            The variational parameters.

        Raises
        ------
        NotImplementedError
            If KL divergence computation is not supported.
        """
        if self.supports_kl:
            return self._kl(var_param0, var_param1)
        raise NotImplementedError()

    def _kl(self, var_param):
        raise NotImplementedError()

    @property
    def supports_kl(self):
        """Whether the approximation family supports closed-form KL divergence computation."""
        return self._supports_kl

    @abstractmethod
    def log_density(self, var_param, x):
        """The log density of the variational distribution.

        Parameters
        ----------
        var_param : `numpy.ndarray`, shape (var_param_dim,)
            The variational parameter.
        x : `numpy.ndarray`, shape (dim,)
            Value at which to evaluate the density."""
        pass

    @abstractmethod
    def mean_and_cov(self, var_param):
        """The mean and covariance of the variational distribution.

        Parameters
        ----------
        var_param : `numpy.ndarray`, shape (var_param_dim,)
            The variational parameter.
        """
        pass

    def pth_moment(self, var_param, p):
        """The absolute pth moment of the variational distribution.

        The absolute pth moment is given by :math:`\\mathbb{E}[|X|^p]`.

        Parameters
        ----------
        var_param : `numpy.ndarray`, shape (var_param_dim,)
            The variational parameter.
        p : `int`

        Raises
        ------
        ValueError
            If `p` value not supported"""
        if self.supports_pth_moment(p):
            return self._pth_moment(var_param, p)
        raise ValueError('p = {} is not a supported moment'.format(p))

    @abstractmethod
    def _pth_moment(self, var_param, p):
        """Get pth moment of the approximating distribution"""
        pass

    @abstractmethod
    def supports_pth_moment(self, p):
        """Whether analytically computing the pth moment is supported"""
        pass

    @property
    def dim(self):
        """Dimension of the space the distribution is defined on"""
        return self._dim

    @property
    def var_param_dim(self):
        """Dimension of the variational parameter"""
        return self._var_param_dim


def _get_mu_log_sigma_pattern(dim):
    ms_pattern = PatternDict(free_default=True)
    ms_pattern['mu'] = NumericVectorPattern(length=dim)
    ms_pattern['log_sigma'] = NumericVectorPattern(length=dim)
    return ms_pattern


class MFGaussian(ApproximationFamily):
    """A mean-field Gaussian approximation family."""
    def __init__(self, dim, seed=1):
        """Create mean field Gaussian approximation family.

        Parameters
        ----------
        dim : `int`
            dimension of the underlying parameter space
        """
        self._rs = npr.RandomState(seed)
        self._pattern = _get_mu_log_sigma_pattern(dim)
        super().__init__(dim, self._pattern.flat_length(True), True, True)

    def init_param(self):
        init_param_dict = dict(mu=np.zeros(self.dim),
                               log_sigma=2*np.ones(self.dim))
        return self._pattern.flatten(init_param_dict)

    def sample(self, var_param, n_samples, seed=None):
        my_rs = self._rs if seed is None else npr.RandomState(seed)
        param_dict = self._pattern.fold(var_param)
        return param_dict['mu'] +  np.exp(param_dict['log_sigma']) * my_rs.randn(n_samples, self.dim)

    def _entropy(self, var_param):
        param_dict = self._pattern.fold(var_param)
        return 0.5 * self.dim * (1.0 + np.log(2*np.pi)) + np.sum(param_dict['log_sigma'])

    def _kl(self, var_param0, var_param1):
        param_dict0 = self._pattern.fold(var_param0)
        param_dict1 = self._pattern.fold(var_param1)
        mean_diff = param_dict0['mu'] - param_dict1['mu']
        log_stdev_diff = param_dict0['log_sigma'] - param_dict1['log_sigma']
        return .5 * np.sum(  np.exp(2*log_stdev_diff)
                           + mean_diff**2 / np.exp(2*param_dict1['log_sigma'])
                           - 2*log_stdev_diff
                           - 1)

    def log_density(self, var_param, x):
        param_dict = self._pattern.fold(var_param)
        return mvn.logpdf(x, param_dict['mu'], np.diag(np.exp(2*param_dict['log_sigma'])))

    def mean_and_cov(self, var_param):
        param_dict = self._pattern.fold(var_param)
        return param_dict['mu'], np.diag(np.exp(2*param_dict['log_sigma']))

    def _pth_moment(self, var_param, p):
        param_dict = self._pattern.fold(var_param)
        vars = np.exp(2*param_dict['log_sigma'])
        if p == 2:
            return np.sum(vars)
        else:  # p == 4
            return 2*np.sum(vars**2) + np.sum(vars)**2

    def supports_pth_moment(self, p):
        return p in [2, 4]


class MFStudentT(ApproximationFamily):
    """A mean-field Student's t approximation family."""
    def __init__(self, dim, df, seed=1):
        if df <= 2:
            raise ValueError('df must be greater than 2')
        self._df = df
        self._rs = npr.RandomState(seed)
        self._pattern = _get_mu_log_sigma_pattern(dim)
        super().__init__(dim, self._pattern.flat_length(True), True, False)

    def init_param(self):
        init_param_dict = dict(mu=np.zeros(self.dim),
                               log_sigma=2*np.ones(self.dim))
        return self._pattern.flatten(init_param_dict)

    def sample(self, var_param, n_samples, seed=None):
        my_rs = self._rs if seed is None else npr.RandomState(seed)
        param_dict = self._pattern.fold(var_param)
        return param_dict['mu'] + np.exp(param_dict['log_sigma']) * my_rs.standard_t(self.df, size=(n_samples, self.dim))

    def entropy(self, var_param):
        # ignore terms that depend only on df
        param_dict = self._pattern.fold(var_param)
        return np.sum(param_dict['log_sigma'])

    def log_density(self, var_param, x):
        if x.ndim == 1:
            x = x[np.newaxis,:]
        param_dict = self._pattern.fold(var_param)
        return np.sum(t_dist.logpdf(x, self.df, param_dict['mu'], np.exp(param_dict['log_sigma'])), axis=-1)

    def mean_and_cov(self, var_param):
        param_dict = self._pattern.fold(var_param)
        df = self.df
        cov = df / (df - 2) * np.diag(np.exp(2*param_dict['log_sigma']))
        return param_dict['mu'], cov

    def _pth_moment(self, var_param, p):
        df = self.df
        if df <= p:
            raise ValueError('df must be greater than p')
        param_dict = self._pattern.fold(var_param)
        scales = np.exp(param_dict['log_sigma'])
        c = df / (df - 2)
        if p == 2:
            return c*np.sum(scales**2)
        else:  # p == 4
            return c**2*(2*(df-1)/(df-4)*np.sum(scales**4) + np.sum(scales**2)**2)

    def supports_pth_moment(self, p):
        return p in [2, 4] and p < self.df

    @property
    def df(self):
        """Degrees of freedom."""
        return self._df


def _get_mu_sigma_pattern(dim):
    ms_pattern = PatternDict(free_default=True)
    ms_pattern['mu'] = NumericVectorPattern(length=dim)
    ms_pattern['Sigma'] = PSDSymmetricMatrixPattern(size=dim)
    return ms_pattern


class MultivariateT(ApproximationFamily):
    """A full-rank multivariate t approximation family."""
    def __init__(self, dim, df, seed=1):
        if df <= 2:
            raise ValueError('df must be greater than 2')
        self._df = df
        self._rs = npr.RandomState(seed)
        self._pattern = _get_mu_sigma_pattern(dim)
        self._log_density = FlattenFunctionInput(
            lambda param_dict, x: multivariate_t_logpdf(x, param_dict['mu'], param_dict['Sigma'], df),
            patterns=self._pattern, free=True, argnums=0)
        super().__init__(dim, self._pattern.flat_length(True), True, False)

    def init_param(self):
        init_param_dict = dict(mu=np.zeros(self.dim),
                               Sigma=10*np.eye(self.dim))
        return self._pattern.flatten(init_param_dict)

    def sample(self, var_param, n_samples, seed=None):
        my_rs = self._rs if seed is None else npr.RandomState(seed)
        df = self.df
        s = np.sqrt(my_rs.chisquare(df, n_samples) / df)
        param_dict = self._pattern.fold(var_param)
        z = my_rs.randn(n_samples, self.dim)
        sqrtSigma = sqrtm(param_dict['Sigma'])
        return param_dict['mu'] + np.dot(z, sqrtSigma)/s[:,np.newaxis]

    def entropy(self, var_param):
        # ignore terms that depend only on df
        param_dict = self._pattern.fold(var_param)
        return .5*np.log(np.linalg.det(param_dict['Sigma']))

    def log_density(self, var_param, x):
        return self._log_density(var_param, x)

    def mean_and_cov(self, var_param):
        param_dict = self._pattern.fold(var_param)
        df = self.df
        return param_dict['mu'], df / (df - 2.) * param_dict['Sigma']

    def _pth_moment(self, var_param, p):
        df = self.df
        if df <= p:
            raise ValueError('df must be greater than p')
        param_dict = self._pattern.fold(var_param)
        sq_scales = np.linalg.eigvalsh(param_dict['Sigma'])
        c = df / (df - 2)
        if p == 2:
            return c*np.sum(sq_scales)
        else:  # p == 4
            return c**2*(2*(df-1)/(df-4)*np.sum(sq_scales**2) + np.sum(sq_scales)**2)

    def supports_pth_moment(self, p):
        return p in [2, 4] and p < self.df

    @property
    def df(self):
        """Degrees of freedom."""
        return self._df



import autograd.numpy as np


def vectorize_if_needed(f, a, axis=-1):
    if a.ndim > 1:
        return np.apply_along_axis(f, axis, a)
    else:
        return f(a)


def ensure_2d(a):
    if a.ndim == 0:
        return a
    while a.ndim < 2:
        a = a[:,np.newaxis]
    return a



import numpy as np

from viabel.approximations import MFGaussian
from viabel.diagnostics import all_diagnostics
from viabel.models import Model, StanModel
from viabel.objectives import ExclusiveKL
from viabel.optimization import adagrad_optimize
from viabel._psis import psislw

all = [
    'bbvi',
    'vi_diagnostics',
]


def bbvi(dimension, *, n_iters=10000, num_mc_samples=10, log_density=None, approx=None, objective=None, fit=None, init_var_param=None, **kwargs):
    """Fit a model using black-box variational inference.

    Currently the objective is optimized using ``viabel.optimization.adagrad_optimize``.

    Parameters
    ----------
    dimension : `int`
        Dimension of the model parameter.
    n_iters : `int`
        Number of iterations of the optimization.
    num_mc_samples : `int`
        Number of Monte Carlo samples to use for estimating the gradient of
        the objective.
    log_density : `function`
        (Unnormalized) log density of the model. Must support automatic
        differentiation with ``autograd``. Either ``log_density`` or ``fit``
        must be provided.
    approx : `ApproximationFamily` object
        The approximation family. The default is to use ``viabel.approximations.MFGaussian``.
    objective : `VariationalObjective` class
        The default is to use ``viabel.objectives.ExclusiveKL``.
    fit : `StanFit4model` object
        If provided, a ``StanModel`` will be used. Both ``fit`` and
        ``log_density`` cannot be given.
    init_var_param
        Initial variational parameter.
    **kwargs
        Keyword arguments to pass to ``adagrad_optimize``.

    Returns
    -------
    results : `dict`
        Contains the following entries: `var_param`, `var_param_history`,
        `objective`
    """
    if objective is not None:
        if fit is not None or log_density is not None or approx is not None:
            raise ValueError('if objective is specified, cannot specify fit, log_density, or approx')
        approx = objective.approx
        model = objective.model
    else:
        if log_density is None:
            if fit is None:
                raise ValueError('either log_density or fit must be specified if objective not given')
            model = StanModel(fit)
        elif fit is None:
            model = Model(log_density)
        else:
            raise ValueError('log_density and fit cannot both be specified')
        if approx is None:
            approx = MFGaussian(dimension)
        objective = ExclusiveKL(approx, model, num_mc_samples)
    if init_var_param is None:
        init_var_param = approx.init_param()
    var_param, var_param_history, _, _ = adagrad_optimize(n_iters,
                                                          objective,
                                                          init_var_param,
                                                          **kwargs)
    results = dict(var_param=var_param,
                   var_param_history=var_param_history,
                   objective=objective)
    return results


def vi_diagnostics(var_param, *, objective=None, model=None, approx=None, n_samples=100000):
    """Check variational inference diagnostics.

    Check Pareto k and 2-divergence diagnostics. Return additional diagnostics
    with mean, standard deviation, and covariance error bounds.

    Parameters
    ----------
    var_param : `numpy.ndarray`, shape (var_param_dim,)
        The variational parameter.
    objective : `function`
    model : `Model` object
    approx : `ApproximationFamily` object
    n_samples : `int`
        The number of samples to use for the diagnostics.

    Returns
    -------
    diagnostics : `dict`
        Also includes samples and smoothed log weights.

    See Also
    --------
    diagostics.all_diagnostics : Compute all diagnostics.
    """
    if objective is None:
        if model is None or approx is None:
            raise ValueError('either objective or both model and approx must be specified')
    elif model is not None or approx is not None:
        raise ValueError('model and/or approx cannot be specified if objective is')
    else:
        model = objective.model
        approx = objective.approx
    if n_samples <= 0:
        raise ValueError('n_samples must be positive')

    return _vi_diagnostics(var_param, model, approx, n_samples)


def _vi_diagnostics(var_param, model, approx, n_samples):
    # first check Pareto k-hat
    samples, smoothed_log_weights, khat = psis_correction(var_param, model, approx, n_samples)
    results = dict(samples=samples,
                   smoothed_log_weights=smoothed_log_weights,
                   khat=khat)
    print('Pareto k is estimated to be khat = {:.2f}'.format(results['khat']))
    if results['khat'] > 0.7:
        print('WARNING: khat > 0.7 means importance sampling is not feasible.')
        print('WARNING: not running further diagnostics')
        return results
    print()
    # if k-hat looks good, check other diagnostics
    if approx.supports_pth_moment(2) and approx.supports_pth_moment(4):
        moment_bound_fn = lambda p: approx.pth_moment(var_param, p)
    else:
        moment_bound_fn = None
    _, q_var = approx.mean_and_cov(var_param)
    results.update(all_diagnostics(smoothed_log_weights,
                                   samples=samples,
                                   moment_bound_fn=moment_bound_fn,
                                   q_var=q_var))
    print('The 2-divergence is estimated to be d2 = {:.2g}'.format(results['d2']))
    if results['d2'] > 4.6: # pragma: no cover
        print('WARNING: d2 > 4.6 means the approximation is very inaccurate')
    elif results['d2'] > 0.1:
        print('WARNING: 0.1 > d2 < 4.6 means the approximation is somewhat '
              'inaccurate. Use importance sampling to decrease error.')
    else:
        print('\nAll diagnostics pass.')
    return results


def psis_correction(var_param, model, approx, n_samples):
    samples, log_weights = samples_and_log_weights(var_param, model, approx, n_samples)
    smoothed_log_weights, khat = psislw(log_weights, overwrite_lw=True)
    return samples.T, smoothed_log_weights, khat


def samples_and_log_weights(var_param, model, approx, n_samples):
    samples = approx.sample(var_param, n_samples)
    log_weights = model(samples) - approx.log_density(var_param, samples)
    return samples, log_weights
